"""StaticPollerManager module

    This module contains a simple Manager which only writes
    inventory chunks on different files for the pollers
    and start them up
"""
import asyncio
import copy
import logging
import os
import shlex
from asyncio.subprocess import Process
from collections import defaultdict
from pathlib import Path
from typing import Dict, List

import aiofiles
import yaml
from cryptography.fernet import Fernet
from suzieq.poller.controller.inventory_async_plugin import \
    InventoryAsyncPlugin
from suzieq.poller.controller.manager.base_manager import Manager
from suzieq.poller.controller.utils.proc_utils import monitor_process
from suzieq.poller.worker.coalescer_launcher import CoalescerLauncher
from suzieq.shared.exceptions import PollingError, SqPollerConfError
from suzieq.shared.utils import get_sq_install_dir

logger = logging.getLogger(__name__)


class StaticManager(Manager, InventoryAsyncPlugin):
    """The StaticPollerManager writes the inventory chunks on files

    The number of pollers is defined in the configuration file with
    the path for inventory files
    """

    def __init__(self, config_data: Dict = None, validate: bool = True):

        super().__init__(config_data, validate)

        self._workers_count = config_data.get("workers", 1)

        # Workers we are already monitoring
        self._running_workers = defaultdict(None)

        # Workers we do not monitor yet
        self._waiting_workers = defaultdict(None)

        # The currently applied chunks
        self._active_chunks = []
        self._poller_tasks_ready = asyncio.Event()

        # Get the running mode
        self._input_dir = config_data.get('input-dir', None)
        self._single_run_mode = config_data.get('single-run-mode', None)
        self._no_coalescer = config_data.get('no-coalescer', False)

        if not self._no_coalescer:
            self._coalescer_launcher = CoalescerLauncher(
                config_data['config'],
                config_data['config-dict']
            )

        # Configure the encyrption of the credential file
        if not self._input_dir:
            cred_key = Fernet.generate_key()
            self._encryptor = Fernet(cred_key)
            # Save the key into an env. variable
            os.environ['SQ_CONTROLLER_POLLER_CRED'] = cred_key.decode()

            # Configure the output directory for the inventory files
            path_name = f'/tmp/.suzieq/inventory.{os.getpid()}'
            self._inventory_path = Path(path_name).resolve()

            try:
                self._inventory_path.mkdir(parents=True, exist_ok=True)
            except FileExistsError:
                raise SqPollerConfError(
                    f'The inventory dir is not a directory: '
                    f'{self._inventory_path}'
                )

            os.environ['SQ_INVENTORY_PATH'] = str(self._inventory_path)

            self._inventory_file_name = 'inv'

        # Define poller parameters
        allowed_args = ['run-once', 'exclude-services',
                        'outputs', 'output-dir', 'service-only',
                        'ssh-config-file', 'config', 'input-dir']

        sq_path = get_sq_install_dir()
        self._args_to_pass = [f'{sq_path}/poller/worker/sq_worker.py']
        for arg, value in config_data.items():
            if arg in allowed_args and value:
                val_list = value if isinstance(value, list) else [value]
                self._args_to_pass.append(f'--{arg}', )
                # All the arguments should be string
                self._args_to_pass += [str(v) for v in val_list]

    @classmethod
    def get_data_model(cls):
        """This is only temporary. In future release I will add mananger
        validation via pydantic
        """
        raise NotImplementedError

    async def apply(self, inventory_chunks: List[Dict]):
        """Apply the inventory chyunks to the pollers

        Args:
            inventory_chunks (List[Dict]): input inventory chunks
        """

        tasks = []
        # In order to prevent any duplicate devices across the pollers, we
        # need to first stop all the pollers involved on the inventory change,
        # and only after all terminated, it is possible to restart them with
        # the new inventory
        pollers_to_launch = []
        n_chunks = len(inventory_chunks)
        if n_chunks != self._workers_count:
            raise PollingError(
                'The number of chunks is different than the number of workers'
            )

        if not self._active_chunks:
            pollers_to_launch = [*range(n_chunks)]
        else:
            for i, chunk in enumerate(inventory_chunks):
                if chunk != self._active_chunks[i]:
                    logger.info(f'Updating worker {i} chunk')
                    pollers_to_launch.append(i)

        # Create the inventory chunks and stop the pollers
        if pollers_to_launch:
            self._poller_tasks_ready.clear()

            logger.info(f'Writing inventory chunks {pollers_to_launch}')
            tasks = [self._write_chunk(i, inventory_chunks[i])
                     for i in pollers_to_launch]
            if self._active_chunks:
                tasks += [self._stop_poller(i) for i in pollers_to_launch]

            # Write the chunks and stop the pollers
            await asyncio.gather(*tasks)

            # Launch all the pollers we need to launch
            # If the mode is debug we do not need to launch the pollers
            if self._single_run_mode != 'debug':
                launch_tasks = [self._launch_poller(i)
                                for i in pollers_to_launch]
                res = await asyncio.gather(
                    *launch_tasks, return_exceptions=True)
                # Check if there are exceptions
                for r in res:
                    if r is not None and isinstance(r, Exception):
                        raise r
                # Copy the active inventory chunks in order to be able
                # to detect any following modifications
                self._active_chunks = copy.deepcopy(inventory_chunks)
            else:
                # When the debug mode is enabled print the instructions to
                # manually launch the workers
                self._print_poller_launch_instructions(n_chunks)
            self._poller_tasks_ready.set()

    async def launch_with_dir(self):
        """Launch a single poller writing the content of and input directory
        produced with the run-once=gather mode
        """
        if self._single_run_mode == 'debug':
            self._print_poller_launch_instructions()
        else:
            await self._launch_poller(0)
        self._poller_tasks_ready.set()

    def get_n_workers(self, _) -> int:
        """returns the content of self._workers_count statically loaded from
           the configuration file

        Attention: This function doesn't use the inventory

        Args:
            inventory (dict, optional): The global inventory.

        Returns:
            int: number of desired workers configured in the configuration
                 file
        """
        return self._workers_count

    async def _execute(self):
        poller_wait_tasks = {}
        tasks = []
        coalescer_task = None
        await self._poller_tasks_ready.wait()

        # Check if we need to start the coalescer
        if not self._no_coalescer:
            coalescer_task = asyncio.create_task(
                self._coalescer_launcher.start_and_monitor_coalescer()
            )
            tasks.append(coalescer_task)

        while self._waiting_workers or self._running_workers:
            if self._waiting_workers:
                # The list of tasks might contain some already terminated
                # workers, remove them before proceeding
                dead_workers = [w for k, w in poller_wait_tasks.items()
                                if k in self._waiting_workers]
                tasks = [t for t in tasks if t not in dead_workers]

                self._running_workers.update(self._waiting_workers)
                new_ptasks = {i: asyncio.create_task(
                    monitor_process(p, f'WORKER {i}'))
                    for i, p in self._waiting_workers.items()}
                poller_wait_tasks.update(new_ptasks)
                tasks += list(new_ptasks.values())
                self._waiting_workers = {}
            # Wait for the tasks
            done, pending = await asyncio.wait(
                tasks,
                return_when=asyncio.FIRST_COMPLETED
            )
            await self._poller_tasks_ready.wait()

            # Check if someone died and investigate why
            for d in done:
                if not self._no_coalescer and d == coalescer_task:
                    # Coalescer died
                    raise PollingError('Unexpected coalescer death')
                # Search for the poller who died
                poller_id = -1
                for i, p in poller_wait_tasks.items():
                    if p == d:
                        poller_id = i
                        break
                if poller_id >= 0:
                    if poller_id not in self._waiting_workers:
                        # Probably unexpected poller died
                        process = self._running_workers[poller_id]

                        if self._single_run_mode and process.returncode == 0:
                            # Worker natural death
                            del self._running_workers[poller_id]
                        else:
                            raise PollingError(f'Unexpected worker {poller_id}'
                                               f' death')
                else:
                    # Someone else died
                    raise PollingError('Unexpected task death')

            tasks = list(pending)

    async def _write_chunk(self, poller_id: int, chunk: Dict):
        """Write the chunk into an output file

        Args:
            id (int): id of the inventory chunk
            chunk (Dict): chunk of the inventory containing the dictionary
        """
        confidential_data = ['password', 'passphrase',
                             'ssh_keyfile', 'jump_host_key_file',
                             'enable_password']
        out_name = {}
        out_name['inv'] = (f'{str(self._inventory_path)}/'
                           f'{self._inventory_file_name}_{poller_id}.yml')
        out_name['cred'] = (f'{str(self._inventory_path)}/'
                            f'cred_{poller_id}')

        inventory_dict = {
            i: {k: v[k] for k in v if k not in confidential_data}
            for i, v in chunk.items()
        }

        inv_data = yaml.safe_dump(inventory_dict)
        if inv_data is None:
            raise PollingError(
                f'Unable to generate inventory file for worker {poller_id}'
            )
        async with aiofiles.open(out_name['inv'], "w") as out_file:
            await out_file.write(inv_data)

        credential_dict = {
            i: {k: v[k] for k in v if k in confidential_data}
            for i, v in chunk.items()
        }

        cred_data = yaml.safe_dump(credential_dict)
        if inv_data is None:
            raise PollingError(
                f'Unable to generate credential file for worker {poller_id}'
            )
        # Encrypt credential data
        enc_cred_data = self._encryptor.encrypt(cred_data.encode('utf-8'))
        async with aiofiles.open(out_name['cred'], "w") as out_file:
            await out_file.write(enc_cred_data.decode())

    async def _launch_poller(self, poller_id: int):
        """Launch a poller with the provided id and chunk, if a poller with
        the given id is already running, stop it and launch it again

        Args:
            id (int): id of the running poller
        """
        curr_args = [*self._args_to_pass, '--worker-id', str(poller_id)]

        # Launch the process
        process = await asyncio.create_subprocess_exec(
            *curr_args,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE
        )
        if not process:
            raise PollingError('Unable to start the poller process')
        self._waiting_workers[poller_id] = process

    def _print_poller_launch_instructions(self, n_chunks: int = 1):
        """Print the commands to execute in order to manually launch
        the poller workers for debugging purpose.

        Args:
            n_chunks (int): the number of chunks the inventory is splitted into
        """
        quoted_args = [shlex.quote(v) for v in self._args_to_pass]
        args_str = ' '.join(quoted_args)
        if self._input_dir:
            print('You can manually start the worker with the following '
                  'command:')
            print()  # Print empty line
            print(f'python {args_str}')
        else:
            print(f'{n_chunks} inventory chunks generated, you can '
                  'now manually start the workers with the following '
                  'commands:')

            for i in range(n_chunks):
                print()  # Print an empty line
                env_to_print = ['SQ_CONTROLLER_POLLER_CRED',
                                'SQ_INVENTORY_PATH']
                for e in env_to_print:
                    print(f'export {e}={os.environ[e]}')

                args_to_print = args_str
                if i > 0:
                    args_to_print += f' --worker-id {i}'
                print(f'python {args_to_print}')

    async def _stop_poller(self, poller_id: int):
        """Kill the poller with the given id. The function first calls a
        SIGTERM, if the process do not exit after 5 seconds, a SIGKILL is sent.

        Args:
            poller_id (int): the poller id
        """
        current_poller = self._running_workers[poller_id]
        await self._stop_process(current_poller)

    async def _stop_process(self, process: Process):
        """Stop a process

        Args:
            process (Process): the process to stop
        """
        if process.returncode is None:
            process.terminate()
            try:
                await asyncio.wait_for(process.wait(), 5)
            except asyncio.TimeoutError:
                process.kill()

    async def _stop(self):
        # Stop all the processes
        running = list(self._running_workers.values())
        waiting = list(self._waiting_workers.values())
        tasks = [self._stop_process(p)
                 for p in [*running, *waiting]]
        await asyncio.gather(*tasks)



import shlex
from nubia import command
from suzieq.cli.nubia_patch import argument

from suzieq.cli.sqcmds.command import SqCommand
from suzieq.sqobjects.device import DeviceObj


@command("device", help="Act on device data")
@argument("status", description="filter by polling status",)
@argument("os", description="NOS(s), space separated")
@argument("version", description="NOS version(s), space separated")
@argument("vendor", description="Vendor(s), space separated")
@argument("model", description="Model(s), space separated")
class DeviceCmd(SqCommand):
    """Basic device information such as OS, version, model etc."""

    def __init__(
            self,
            engine: str = "",
            hostname: str = "",
            start_time: str = "",
            end_time: str = "",
            view: str = "",
            namespace: str = "",
            format: str = "",  # pylint: disable=redefined-builtin
            query_str: str = " ",
            columns: str = "default",
            os: str = '',
            version: str = '',
            status: str = '',
            vendor: str = '',
            model: str = '',
    ) -> None:
        super().__init__(
            engine=engine,
            hostname=hostname,
            start_time=start_time,
            end_time=end_time,
            view=view,
            namespace=namespace,
            columns=columns,
            format=format,
            query_str=query_str,
            sqobj=DeviceObj,
        )

        # Nubia eats up the first and last quote. Detect and add back
        model = model.strip()
        if model.count("'") % 2 != 0:
            if not model.startswith("'"):
                model = f"'{model}"
            elif not model.endswith("'"):
                model = f"{model}'"
        if model:
            model = shlex.split(model)

        self.lvars = {
            'os': os.split(),
            'version': version.split(),
            'status': status.split(),
            'model': model,
            'vendor': vendor.split()
        }



'''
Serve up a fake SSH server acting as device CLI
'''
import sys
import argparse
from pathlib import Path
import crypt

import asyncio
import asyncssh

from suzieq.shared.utils import get_sq_install_dir


class MySSHServerSession(asyncssh.SSHServerSession):
    '''Customised ssh server session for serving precanned outputs'''

    def __init__(self, input_dir: str = None):
        self._input = ''
        self._data = None
        self.input_dir = Path(f'{input_dir}')
        self.run_as_shell = False
        self.prompt = '# '
        self._status = 0
        self._chan = None

    def connection_made(self, chan):
        '''callback when connection is estd'''
        self._chan = chan

    def shell_requested(self):
        '''returns True if ssh client requested shell'''
        self.run_as_shell = True
        return True

    def eof_received(self):
        '''EOF received, end'''
        self._chan.write('Goodbye!\n')
        self._chan.exit(0)

    def get_cmd_file(self, command: str, fmt: str = '.txt') -> str:
        '''Return the file containing the output of the requested cmd'''

        if command:
            if command == 'cat /proc/uptime; hostnamectl; show version':
                return f'{self.input_dir}/nos.txt'

            cmdfile = command.replace("*", "all").replace(" ", "_") \
                .replace("-", "_")

            filepath = self.input_dir / f'{cmdfile}{fmt}'

            if filepath.exists():
                return filepath

            print(f'No file {filepath} found')
            return ''

        return ''

    def _exec_cmd(self, command):
        '''The routine to execute command and return data'''
        data = 'Command not found\n'
        if command in ['exit', 'quit']:
            self.eof_received()
            return ''
        if 'json' in command:
            command = command.split('|')[0].strip()
            fmt = '.json'
        else:
            fmt = '.txt'
        cmdfile = self.get_cmd_file(command, fmt)
        if cmdfile:
            with open(cmdfile, 'r', encoding='utf8') as fhndl:
                data = fhndl.read()
            self._status = 0
        else:
            self._status = -1
        return data

    def exec_requested(self, command):
        '''Return the data for the specified command is possible'''

        self.run_as_shell = False
        command = command.rstrip('\n')
        self._data = self._exec_cmd(command)
        return True

    def data_received(self, data, datatype):
        '''Shell handler'''

        command = data.rstrip('\n')
        indata = self._exec_cmd(command)
        self._chan.write(indata)
        self._chan.write(self.prompt)

    def session_started(self):
        if self._status == 0:
            if self._data is not None:
                self._chan.write(self._data)
            else:
                self._chan.write(f'{self.prompt}')
            if not self.run_as_shell:
                self._chan.exit(0)
        elif not self.run_as_shell:
            self._chan.exit(1)
        elif self.run_as_shell:
            self._chan.write('Command not found\n')
            self._chan.write(f'{self.prompt}')


class MySSHServer(asyncssh.SSHServer):
    '''My own customized SSH server class to serve precanned outputs'''

    def __init__(self, input_dir: str = ''):
        self.passwords = {'vagrant': 'vaqRzE48Dulhs'}   # password of 'vagrant'
        self.input_dir = input_dir
        if not Path(f'{input_dir}').exists():
            print('ERROR: Received non-existent files dir '
                  f'{input_dir}')

    def connection_made(self, conn):
        '''Callback when connection is estd'''
        print('SSH connection received from '
              f'{conn.get_extra_info("peername")[0]}')

    # pylint: disable=redefined-outer-name
    def connection_lost(self, exc):
        '''Callback on losing client connection'''
        if exc:
            print('SSH connection error: ' + str(exc), file=sys.stderr)
        else:
            print('SSH connection closed.')

    def begin_auth(self, username):
        '''Start authorization'''
        # If the user's password is the empty string, no auth is required
        return self.passwords.get(username) != ''

    def password_auth_supported(self):
        '''Do we support password-based auth'''
        return True

    def validate_password(self, username, password):
        '''check if password is correct'''
        passwd = self.passwords.get(username, '*')
        return crypt.crypt(password, passwd) == passwd

    def session_requested(self):
        '''return new SSH session'''
        return MySSHServerSession(input_dir=self.input_dir)


async def start_server(port=10000, input_dir: str = None):
    """Run sim ssh server using the files in the given dir"""

    keyfile = Path(get_sq_install_dir()) / 'config/etc' / 'ssh_insecure_key'
    await asyncssh.listen(
        '127.0.0.1', port,
        server_factory=lambda: MySSHServer(input_dir=input_dir),
        server_host_keys=[keyfile])

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-p", "--listening-port", type=int, default=10000,
        help="Listening port of the ssh server (default: 10000)")
    parser.add_argument(
        "-d", "--input-dir", type=str, default=None,
        help="Input dir to search for host files")
    args = parser.parse_args()

    loop = asyncio.get_event_loop()

    if not Path(args.input_dir).exists():
        print(f'ERROR: Path {args.input_dir} does not exist, aborting')
        sys.exit(1)

    try:
        loop.run_until_complete(start_server(
            port=args.listening_port, input_dir=args.input_dir)
        )
    except (OSError, asyncssh.Error) as exc:
        sys.exit('Error starting server: ' + str(exc))

    loop.run_forever()



import base64
import os
from enum import Enum
from importlib.util import find_spec

import pandas as pd
import streamlit as st
from IPython.display import Markdown
from streamlit.server.server import Server
try:
    from streamlit.script_run_context import get_script_run_ctx
except ModuleNotFoundError:
    # streamlit < 1.4
    from streamlit.report_thread import (  # type: ignore
        get_report_ctx as get_script_run_ctx,
    )
from suzieq.sqobjects import get_sqobject

SUZIEQ_COLOR = "#68279D"


class SuzieqMainPages(str, Enum):
    '''Pages in Suzieq GUI'''
    STATUS = "Status"
    XPLORE = "Xplore"
    PATH = "Path"
    SEARCH = "Search"


def pandas_df_to_markdown_table(df: pd.DataFrame) -> str:
    """Convert a Dataframe into a markdown table

    Args:
        df (pd.DataFrame): input dataframe

    Returns:
        str: markdown table
    """

    fmt = ['---' for i in range(len(df.columns))]
    df_fmt = pd.DataFrame([fmt], columns=df.columns)
    df_formatted = pd.concat([df_fmt, df])
    return Markdown(df_formatted.to_csv(sep="|", index=False)).data


def display_help_icon(url: str):
    '''Display Help Icon with click to take you to appropriate page'''

    help_img = f'{get_image_dir()}/helps.png'
    with open(help_img, "rb") as f:
        img = base64.b64encode(f.read()).decode()

    st.sidebar.markdown(
        f'<a target="_help" href="{url}"><img class="help-img" '
        f'src="data:image/png;base64,'
        f'{img}"></a>',
        unsafe_allow_html=True)


@st.cache(ttl=90, allow_output_mutation=True, show_spinner=False,
          max_entries=20)
def gui_get_df(table: str,
               config_file: str,
               verb: str = 'get', **kwargs) -> pd.DataFrame:
    """Get the cached value of the table provided

    The only verbs supported are get and find.

    Args:
        table ([str]): The table for which to get the data
        config_file ([str]): The config file for the sqobject
        verb (str, optional): . Defaults to 'get'.

    Returns:
        [pandas.DataFrame]: The dataframe

    Raises:
        ValueError: If the verb is not supported
    """
    view = kwargs.pop('view', 'latest')
    columns = kwargs.pop('columns', ['default'])
    stime = kwargs.pop('start_time', '')
    etime = kwargs.pop('end_time', '')

    sqobject = get_sqobject(table)(view=view, start_time=stime, end_time=etime,
                                   config_file=config_file)
    if columns == ['all']:
        columns = ['*']
    fn = getattr(sqobject, verb, None)
    if not fn or not callable(fn):
        raise ValueError(f'Unsupported verb {verb}')

    if verb not in ['assert', 'find']:
        df = fn(columns=columns, **kwargs)
    else:
        df = fn(**kwargs)

    if not df.empty:
        df = sqobject.humanize_fields(df)
        if table == 'address':
            if 'ipAddressList' in df.columns:
                df = df.explode('ipAddressList').fillna('')
            if 'ip6AddressList' in df.columns:
                df = df.explode('ip6AddressList').fillna('')

    if verb == 'summarize':
        return df
    if columns not in [['*'], ['default']]:
        return df[columns].reset_index(drop=True)
    return df.reset_index(drop=True)


def get_base_url():
    '''Return the base URL of the page.
    Usually connections are http://localhost:8501. But the port can change
    or it can be a remote connection. And so, its useful to get the base
    URL for use with links on various pages.
    '''
    session_id = get_script_run_ctx().session_id
    # pylint: disable=protected-access
    session_info = Server.get_current()._get_session_info(session_id)

    if session_info:
        return f'http://{session_info.ws.request.host}/'

    return 'http://localhost:8501/'


def get_session_id():
    '''Return Streamlit's session ID'''
    ctx = get_script_run_ctx()
    if ctx is None:
        raise Exception("Failed to get the thread context")
    return ctx.session_id


def get_main_session_by_id(session_id):
    """This returns the session state based on the ID provided

    This is relying on Streamlit internals and is not a public API.

    Args:
        session_id ([type]): The session id string

    Returns:
        [type]: session state associated with session or None
    """
    # pylint: disable=protected-access
    session = Server.get_current()._session_info_by_id.get(session_id, None)
    if session:
        return session.session.session_state

    return None


def set_def_aggrid_options(grid_options: dict):
    '''Set default grid options for aggrid tables

    Invoke this function by passing it the output pf gb.build
    '''
    grid_options['rowStyle'] = {'background': 'white'}
    grid_options['cacheQuickFilter'] = True
    grid_options['suppressDragLeaveHidesColumns'] = True
    grid_options['enableCellTextSelection'] = True
    grid_options['ensureDomOrder'] = True

    return grid_options


def get_image_dir():
    '''Get directory where images are stored'''
    return(os.path.dirname(find_spec('suzieq.gui')
                           .loader.path) + '/images')


def color_row(row, **kwargs):
    """Color the appropriate column red if the status has failed"""
    fieldval = kwargs.pop("fieldval", "down")
    field = kwargs.pop("field", "state")
    color = kwargs.pop("color", "black")
    bgcolor = kwargs.pop("bgcolor", "yellow")

    if row[field] in fieldval:
        return [f"background-color: {bgcolor}; color: {color}"]*len(row)
    else:
        return [""]*len(row)


def color_element_red(value, **kwargs):
    '''Use with applymap to color a cell based on a value'''
    fieldval = kwargs.pop("fieldval", "down")
    if value not in fieldval:
        return "background-color: red; color: white;"
    # Assuming light theme here
    return "background-color: white; color: black;"


def ifstate_red(row):
    '''Color interface state red if admin state is up, but not oper state'''
    llen = len(row.index.tolist())
    if row.adminState == "up":
        if row.state == "down":
            return ["background-color: red; color: white;"]*llen
        elif row.state != "up":
            return ["color:  gray;"]*llen

    return ["background-color: white; color: black;"]*llen


def sq_gui_style(df, table, is_assert=False):
    """Apply appropriate styling for the dataframe based on the table"""

    if is_assert:
        if not df.empty:
            return df.style.apply(color_row, axis=1, field='result',
                                  fieldval=['fail'], bgcolor='darkred',
                                  color='white')
        else:
            return df

    if table == 'bgp' and 'state' in df.columns:
        return df.style.hide_index() \
                       .applymap(color_element_red, fieldval=['Established'],
                                 subset=pd.IndexSlice[:, ['state']])
    elif table == 'ospf' and 'adjState' in df.columns:
        return df.style.hide_index() \
            .applymap(color_element_red,
                      fieldval=["full", "passive"],
                      subset=pd.IndexSlice[:, ['adjState']])

    elif table == "routes" and 'prefix' in df.columns:
        return df.style.hide_index() \
            .apply(color_row, axis=1, fieldval=['0.0.0.0/0'],
                   field='prefix')
    elif table == "interfaces" and 'state' in df.columns:
        return df.style.hide_index().apply(ifstate_red, axis=1)
    elif table == "device":
        return df.style.hide_index() \
            .apply(color_row, axis=1, fieldval=['dead', 'neverpoll'],
                   field='status', bgcolor='red', color='white')
    else:
        return df.style.hide_index()


def get_color_styles(color: str) -> str:
    """Compile some hacky CSS to override the theme color."""
    # fmt: off
    color_selectors = ["a", "a:hover", "*:not(textarea).st-ex:hover",
                       ".st-en:hover"]
    bg_selectors = [".st-da", "*:not(button).st-en:hover"]
    border_selectors = [".st-ft", ".st-fs", ".st-fr", ".st-fq", ".st-ex:hover",
                        ".st-en:hover"]
    # fmt: on
    css_root = "#root { --primary: %s }" % color
    css_color = ", ".join(color_selectors) + "{ color: %s !important }" % color
    css_bg = ", ".join(bg_selectors) + \
        "{ background-color: %s !important }" % color
    css_border = ", ".join(border_selectors) + \
        "{ border-color: %s !important }" % color
    other = ".decoration { background: %s !important }" % color
    return f"<style>{css_root}{css_color}{css_bg}{css_border}{other}</style>"



from dataclasses import dataclass
from typing import List

import numpy as np
import streamlit as st
import altair as alt

from suzieq.gui.stlit.guiutils import gui_get_df, SuzieqMainPages
from suzieq.gui.stlit.pagecls import SqGuiPage


@dataclass
class StatusSessionState:
    '''Session state for status page'''
    page: str = SuzieqMainPages.STATUS.value
    namespace: str = ''


class StatusPage(SqGuiPage):
    '''Page for Main status page'''
    _title: str = SuzieqMainPages.STATUS.value
    _state: StatusSessionState = StatusSessionState()
    _config_file: str = st.session_state.get('config_file', '')

    @property
    def add_to_menu(self):
        return True

    def build(self):
        self._get_state_from_url()
        self._create_sidebar()
        layout = self._create_layout()
        self._render(layout)

    def _create_sidebar(self) -> None:
        '''Draw appropriate sidebar for the page'''

        do_refresh = st.sidebar.button('Refresh')

        devdf = gui_get_df('device', self._config_file,
                           columns=['namespace', 'hostname'])
        if devdf.empty:
            st.error('Unable to retrieve any namespace info')
            st.stop()

        namespaces = [''] + sorted(devdf.namespace.unique().tolist())
        if self._state.namespace:
            nsidx = namespaces.index(self._state.namespace)
        else:
            nsidx = 0
        namespace = st.sidebar.selectbox('Namespace', namespaces, index=nsidx,
                                         key='status_ns',
                                         on_change=self._sync_state)

        if do_refresh:
            st.caching.clear_memo_cache()

        st.sidebar.markdown(
            '''This page provides an overview of the overall network status

Select one of the following pages from the Page menu to investigate further.
* __Xplore__: Look at all the data, look at summaries, run asserts, queries
              and more
* __Path__: Trace the paths between destinations in a given namespace
* __Search__: Search for addresses in various tables. Just type in any address
              you want to search. You can specify multiple addresses, space
               separated. See the search page for more help.

__Caching is enabled by default for 90 secs on all pages__. You can clear the
 cache by hitting the refresh button on this page or selecting "Clear Cache"
 option from the drop down menu on the top right hand corner
        ''')

        if namespace != self._state.namespace:
            self._state.namespace = namespace

    def _create_layout(self) -> dict:
        col1, _, col2 = st.columns([2, 1, 2])
        with col1:
            dev_chart = st.empty()
        with col2:
            if_chart = st.empty()
        col3, _, col4 = st.columns([2, 1, 2])
        with col3:
            bgp_chart = st.empty()
        with col4:
            ospf_chart = st.empty()

        poller_ph = st.empty()

        return {
            'dev_chart': dev_chart,
            'if_chart': if_chart,
            'ospf_chart': ospf_chart,
            'bgp_chart': bgp_chart,
            'poller': poller_ph,
        }

    def _render(self, layout) -> None:
        '''Render the page'''
        if self._state.namespace:
            ns_list = [self._state.namespace]
        else:
            ns_list = []

        self._chart_dev(layout, ns_list)
        self._chart_interface(layout, ns_list)
        self._chart_bgp(layout, ns_list)
        self._chart_ospf(layout, ns_list)
        self._chart_poller(layout, ns_list)

    def _sync_state(self) -> None:
        wsstate = st.session_state

        self._state.namespace = wsstate.status_ns
        self._save_page_url()

    def _chart_dev(self, layout: dict, ns_list: List[str]) -> None:
        '''Chart the devices table status'''

        dev_df = gui_get_df('device', self._config_file,
                            namespace=ns_list, columns=['*'])
        if dev_df.empty:
            layout['dev_chart'].warning('No device info found')
            return

        dev_status = dev_df \
            .groupby(by=['namespace', 'status'])['hostname'].count() \
            .reset_index() \
            .rename({'hostname': 'count'}, axis=1)

        dev_chart = alt.Chart(dev_status, title='Devices') \
                       .mark_bar(tooltip=True) \
                       .encode(y='status', x='count:Q', row='namespace',
                               color=alt.Color(
                                   'status',
                                   scale=alt.Scale(
                                       domain=[
                                           'alive', 'dead', 'neverpoll'],
                                       range=['green', 'red', 'darkred']))
                               )
        layout['dev_chart'].altair_chart(dev_chart)

    def _chart_interface(self, layout: dict, ns_list: List[str]) -> None:
        '''Chart the interfaces table status'''

        if_df = gui_get_df('interfaces', self._config_file,
                           namespace=ns_list,
                           columns=['default'])
        if if_df.empty:
            layout['if_chart'].warning('No Interface info found')
            return

        if_df['state'] = np.where((if_df.state == "down") & (
            if_df.adminState == "down"), "adminDown", if_df.state)

        if_status = if_df.groupby(by=['namespace', 'state'])['hostname'] \
                         .count() \
                         .reset_index() \
                         .rename({'hostname': 'count'}, axis=1)

        if_chart = alt \
            .Chart(if_status, title='Interfaces') \
            .mark_bar(tooltip=True) \
            .encode(y='state', x='count:Q', row='namespace',
                    color=alt.Color(
                        'state',
                        scale=alt.Scale(domain=['up', 'errDisabled',
                                                'down', 'notPresent',
                                                'notConnected'],
                                        range=['green', 'darkred',
                                               'red', 'grey', 'darkgray']))
                    )
        layout['if_chart'].altair_chart(if_chart)

    def _chart_bgp(self, layout: dict, ns_list: List[str]) -> None:
        '''Chart the BGP table status'''

        bgp_df = gui_get_df('bgp', self._config_file,
                            namespace=ns_list, columns=['default'])

        if bgp_df.empty:
            return

        bgp_status = bgp_df.groupby(by=['namespace', 'state'])['hostname'] \
                           .count() \
                           .reset_index() \
                           .rename({'hostname': 'count'}, axis=1)

        bgp_chart = alt.Chart(bgp_status, title='BGP') \
                       .mark_bar(tooltip=True) \
                       .encode(y='state', x='count:Q', row='namespace',
                               color=alt.Color(
                                   'state',
                                   scale=alt.Scale(
                                       domain=['Established',
                                               'NotEstd', 'dynamic'],
                                       range=['green', 'red', 'orange']))
                               )
        layout['bgp_chart'].altair_chart(bgp_chart)

    def _chart_ospf(self, layout: dict, ns_list: List[str]) -> None:
        '''Chart OSPF status'''

        ospf_df = gui_get_df('ospf', self._config_file,
                             namespace=ns_list, columns=['default'])
        if ospf_df.empty:
            return

        ospf_df['state'] = np.where(ospf_df.ifState == "adminDown",
                                    "adminDown", ospf_df.adjState)

        ospf_status = ospf_df.groupby(by=['namespace', 'state'])['hostname'] \
                             .count() \
                             .reset_index() \
                             .rename({'hostname': 'count'}, axis=1)

        ospf_chart = alt.Chart(ospf_status, title='OSPF') \
                        .mark_bar(tooltip=True) \
                        .encode(y='state', x='count:Q', row='namespace',
                                color=alt.Color(
                                    'state',
                                    scale=alt.Scale(
                                        domain=['full', 'fail',
                                                'adminDown', 'passive'],
                                        range=['green', 'red', 'orange',
                                               'peach']))
                                )
        layout['ospf_chart'].altair_chart(ospf_chart)

    def _chart_poller(self, layout: dict, ns_list: List[str]) -> None:
        '''Display poller status'''

        sqdf = gui_get_df('sqPoller',
                          self._config_file,
                          columns=['namespace', 'hostname', 'timestamp'],
                          service='device', namespace=ns_list)
        if sqdf.empty:
            st.warning('No Poller information found')
            return

        hosts = sqdf.groupby(by=['namespace'])['hostname'] \
                    .nunique() \
                    .reset_index() \
                    .rename({'hostname': '#devices'}, axis=1)
        times = sqdf.groupby(by=['namespace'])['timestamp'] \
                    .max().reset_index() \
                          .rename({'timestamp': 'lastPolledTime'}, axis=1)
        pstats = times.merge(hosts, on=['namespace'])

        layout['poller'].subheader('Poller Status')
        layout['poller'].dataframe(pstats)



"""
This module contains the logic of the writer for the 'parquet' mode
"""
import logging
import os
from typing import Dict

import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq

from suzieq.shared.exceptions import SqPollerConfError
from suzieq.poller.worker.writers.output_worker import OutputWorker

logger = logging.getLogger(__name__)


class ParquetOutputWorker(OutputWorker):
    """ParquetOutputWorker writes the data retrived by
    the poller in a parquet output directory
    """

    def __init__(self, **kwargs):
        data_directory = kwargs.get('data_dir')

        if not data_directory:
            output_dir = '/tmp/suzieq/parquet-out/'
            logger.warning(
                'No output directory for parquet specified, using '
                '/tmp/suzieq/parquet-out'
            )
        else:
            output_dir = data_directory

        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        if not os.path.isdir(output_dir):
            raise SqPollerConfError(
                f'Output directory {output_dir} is not a directory'
            )

        logger.info(f'Parquet outputs will be under {output_dir}')
        self.root_output_dir = output_dir

    def write_data(self, data: Dict):
        """Write the data into the Parquet output directory

        Args:
            data (Dict): dictionary containing the data to store.
        """
        cdir = os.path.join(self.root_output_dir, data['topic'])
        if not os.path.isdir(cdir):
            os.makedirs(cdir)

        # dtypes = {x: data['schema'].field(x).type.__str__()
        #           if 'list' not in data['schema'].field(x).type.__str__()
        #           else data['schema'].field(x).type.to_pandas_dtype()
        #           for x in data['schema'].names}

        df = pd.DataFrame.from_dict(data['records'])
        # df.to_parquet(
        #     path=cdir,
        #     partition_cols=data['partition_cols'],
        #     index=True,
        #     engine='pyarrow')
        # pq.write_metadata(
        #     self.schema,'{}/_metadata'.format(cdir),
        #     version='2.0',
        #     coerce_timestamps='us')

        table = pa.Table.from_pandas(df, schema=data['schema'],
                                     preserve_index=False)

        pq.write_to_dataset(
            table,
            root_path=cdir,
            partition_cols=data['partition_cols'],
            version='2.0',
            compression='ZSTD',
            row_group_size=100000,
        )



from ipaddress import ip_interface

from suzieq.sqobjects.basicobj import SqObject
from suzieq.shared.utils import validate_macaddr, validate_network


class AddressObj(SqObject):
    '''The object providing access to the address table'''

    def __init__(self, **kwargs):
        super().__init__(table='address', **kwargs)
        self._valid_get_args = ['namespace', 'hostname', 'address', 'prefix',
                                'columns', 'ipvers', 'vrf', 'query_str',
                                'type', 'ifname']
        self._unique_def_column = ['ipAddress']

    def validate_get_input(self, **kwargs):
        if kwargs.get('prefix', []) and kwargs.get('address', []):
            raise AttributeError("Cannot specify address and prefix together")

        if kwargs.get('prefix', []):
            for p in kwargs['prefix']:
                if not validate_network(p):
                    raise ValueError("Invalid prefix specified")

        if kwargs.get('address', []):
            for a in kwargs['address']:
                try:
                    ip_interface(a)
                except ValueError:
                    if not validate_macaddr(a):
                        raise ValueError("Invalid address specified")

        super().validate_get_input(**kwargs)
